import asyncio
import inspect
from abc import ABC, abstractmethod
from types import CoroutineType, GeneratorType
from typing import Any, Dict, List, Union
from weakref import WeakSet

from graphql.execution.executors.asyncio import AsyncioExecutor
from promise import Promise

from .base import BaseConnectionContext, BaseSubscriptionServer

from .constants import GQL_COMPLETE, GQL_CONNECTION_ACK, GQL_CONNECTION_ERROR, GQL_DATA
from .observable_aiter import setup_observable_extension

setup_observable_extension()
CO_ITERABLE_COROUTINE = inspect.CO_ITERABLE_COROUTINE


# Copied from graphql-core v3.1.0 (graphql/pyutils/is_awaitable.py)
def is_awaitable(value: Any) -> bool:
    """Return true if object can be passed to an ``await`` expression.
    Instead of testing if the object is an instance of abc.Awaitable, it checks
    the existence of an `__await__` attribute. This is much faster.
    """
    return (
        # check for coroutine objects
        isinstance(value, CoroutineType)
        # check for old-style generator based coroutine objects
        or isinstance(value, GeneratorType)
        and bool(value.gi_code.co_flags & CO_ITERABLE_COROUTINE)
        # check for other awaitables (e.g. futures)
        or hasattr(value, "__await__")
    )


async def resolve(
    data: Any, _container: Union[List, Dict] = None, _key: Union[str, int] = None
) -> None:
    """
    Recursively wait on any awaitable children of a data element and resolve any
    Promises.
    """
    if is_awaitable(data):
        data = await data
        if isinstance(data, Promise):
            data = data.value  # type: Any
        if _container is not None:
            _container[_key] = data
    if isinstance(data, dict):
        items = data.items()
    elif isinstance(data, list):
        items = enumerate(data)
    else:
        items = None
    if items is not None:
        children = [
            asyncio.ensure_future(resolve(child, _container=data, _key=key))
            for key, child in items
        ]
        if children:
            await asyncio.wait(children)


class BaseAsyncConnectionContext(BaseConnectionContext, ABC):
    def __init__(self, ws, request_context=None):
        super().__init__(ws, request_context=request_context)
        self.pending_tasks = WeakSet()

    @abstractmethod
    async def receive(self):
        raise NotImplementedError("receive method not implemented")

    @abstractmethod
    async def send(self, data):
        ...

    @property
    @abstractmethod
    def closed(self):
        ...

    @abstractmethod
    async def close(self, code):
        ...

    def remember_task(self, task):
        self.pending_tasks.add(task)
        # Clear completed tasks
        self.pending_tasks -= WeakSet(
            task for task in self.pending_tasks if task.done()
        )

    async def unsubscribe(self, op_id):
        async_iterator = super().unsubscribe(op_id)
        if getattr(async_iterator, "future", None) and async_iterator.future.cancel():
            await async_iterator.future

    async def unsubscribe_all(self):
        awaitables = [self.unsubscribe(op_id) for op_id in list(self.operations)]
        for task in self.pending_tasks:
            task.cancel()
            awaitables.append(task)
        if awaitables:
            try:
                await asyncio.gather(*awaitables)
            except asyncio.CancelledError:
                pass


class BaseAsyncSubscriptionServer(BaseSubscriptionServer, ABC):
    graphql_executor = AsyncioExecutor

    def __init__(self, schema, keep_alive=True, loop=None):
        self.loop = loop
        super().__init__(schema, keep_alive)

    @abstractmethod
    async def handle(self, ws, request_context=None):
        ...

    def process_message(self, connection_context, parsed_message):
        task = asyncio.ensure_future(
            super().process_message(connection_context, parsed_message), loop=self.loop
        )
        connection_context.remember_task(task)
        return task

    async def on_open(self, connection_context):
        pass

    async def on_connect(self, connection_context, payload):
        pass

    async def on_connection_init(self, connection_context, op_id, payload):
        try:
            await self.on_connect(connection_context, payload)
            await self.send_message(connection_context, op_type=GQL_CONNECTION_ACK)
        except Exception as e:
            await self.send_error(connection_context, op_id, e, GQL_CONNECTION_ERROR)
            await connection_context.close(1011)

    async def on_start(self, connection_context, op_id, params, data_op_type=GQL_DATA):
        # Attempt to unsubscribe first in case we already have a subscription
        # with this id.
        await connection_context.unsubscribe(op_id)

        execution_result = self.execute(params)

        connection_context.register_operation(op_id, execution_result)
        if hasattr(execution_result, "__aiter__"):
            iterator = await execution_result.__aiter__()
            connection_context.register_operation(op_id, iterator)
            try:
                async for single_result in iterator:
                    if not connection_context.has_operation(op_id):
                        break
                    await self.send_execution_result(
                        connection_context, op_id, single_result, data_op_type
                    )
            except Exception as e:
                await self.send_error(connection_context, op_id, e)
        else:
            try:
                if is_awaitable(execution_result):
                    execution_result = await execution_result
                await self.send_execution_result(
                    connection_context, op_id, execution_result, data_op_type
                )
            except Exception as e:
                await self.send_error(connection_context, op_id, e)
        await self.send_message(connection_context, op_id, GQL_COMPLETE)
        await connection_context.unsubscribe(op_id)
        await self.on_operation_complete(connection_context, op_id)

    async def send_message(
        self, connection_context, op_id=None, op_type=None, payload=None
    ):
        if op_id is None or connection_context.has_operation(op_id):
            message = self.build_message(op_id, op_type, payload)
            return await connection_context.send(message)

    async def on_operation_complete(self, connection_context, op_id):
        pass

    async def send_execution_result(self, connection_context, op_id, execution_result, data_op_type=GQL_DATA):
        # Resolve any pending promises
        await resolve(execution_result.data)
        await super().send_execution_result(connection_context, op_id, execution_result, data_op_type=data_op_type)



from flask import Flask, make_response
from flask_graphql import GraphQLView
from flask_sockets import Sockets

from graphql_ws_apollo.gevent import GeventSubscriptionServer

from schema import schema
from template import render_graphiql

app = Flask(__name__)
app.debug = True
sockets = Sockets(app)


@app.route("/graphiql")
def graphql_view():
    return make_response(render_graphiql())


app.add_url_rule(
    "/graphql", view_func=GraphQLView.as_view("graphql", schema=schema, graphiql=False)
)

subscription_server = GeventSubscriptionServer(schema)
app.app_protocol = lambda environ_path_info: "graphql-ws"


@sockets.route("/subscriptions")
def echo_socket(ws):
    subscription_server.handle(ws)
    return []


if __name__ == "__main__":
    from gevent import pywsgi
    from geventwebsocket.handler import WebSocketHandler

    server = pywsgi.WSGIServer(("", 5000), app, handler_class=WebSocketHandler)
    server.serve_forever()



import graphene
from rx import Observable


class Query(graphene.ObjectType):
    hello = graphene.String()

    def resolve_hello(self, info, **kwargs):
        return "world"


class Subscription(graphene.ObjectType):

    count_seconds = graphene.Int(up_to=graphene.Int())

    def resolve_count_seconds(root, info, up_to=5):
        return (
            Observable.interval(1000)
            .map(lambda i: "{0}".format(i))
            .take_while(lambda i: int(i) <= up_to)
        )


schema = graphene.Schema(query=Query, subscription=Subscription)



from asyncio import Future

from rx.core import Observable
from rx.internal import extensionmethod


async def __aiter__(self):
    source = self

    class AIterator:
        def __init__(self):
            self.notifications = []
            self.future = Future()

            self.disposable = source.materialize().subscribe(self.on_next)

        def __aiter__(self):
            return self

        def dispose(self):
            self.disposable.dispose()

        def feeder(self):
            if not self.notifications or self.future.done():
                return

            notification = self.notifications.pop(0)
            kind = notification.kind
            if kind == "N":
                self.future.set_result(notification.value)
            if kind == "E":
                self.future.set_exception(notification.exception)
            if kind == "C":
                self.future.set_exception(StopAsyncIteration)

        def on_next(self, notification):
            self.notifications.append(notification)
            self.feeder()

        async def __anext__(self):
            self.feeder()

            value = await self.future
            self.future = Future()
            return value

    return AIterator()


def setup_observable_extension():
    extensionmethod(Observable)(__aiter__)



from unittest import mock

import json
import promise

import pytest

from graphql_ws_apollo import base, base_async

pytestmark = pytest.mark.asyncio


class AsyncMock(mock.MagicMock):
    async def __call__(self, *args, **kwargs):
        return super().__call__(*args, **kwargs)


class TstServer(base_async.BaseAsyncSubscriptionServer):
    def handle(self, *args, **kwargs):
        pass  # pragma: no cover


@pytest.fixture
def server():
    return TstServer(schema=None)


async def test_terminate(server: TstServer):
    context = AsyncMock()
    await server.on_connection_terminate(connection_context=context, op_id=1)
    context.close.assert_called_with(1011)


async def test_send_error(server: TstServer):
    context = AsyncMock()
    context.has_operation = mock.Mock()
    await server.send_error(connection_context=context, op_id=1, error="test error")
    context.send.assert_called_with(
        {"id": 1, "type": "error", "payload": {"message": "test error"}}
    )


async def test_message(server):
    server.process_message = AsyncMock()
    context = AsyncMock()
    msg = {"id": 1, "type": base.GQL_CONNECTION_INIT, "payload": ""}
    await server.on_message(context, msg)
    server.process_message.assert_called_with(context, msg)


async def test_message_str(server):
    server.process_message = AsyncMock()
    context = AsyncMock()
    msg = {"id": 1, "type": base.GQL_CONNECTION_INIT, "payload": ""}
    await server.on_message(context, json.dumps(msg))
    server.process_message.assert_called_with(context, msg)


async def test_message_invalid(server):
    server.send_error = AsyncMock()
    await server.on_message(connection_context=None, message="'not-json")
    assert server.send_error.called


async def test_resolver(server):
    server.send_message = AsyncMock()
    result = mock.Mock()
    result.data = {"test": [1, 2]}
    result.errors = None
    await server.send_execution_result(
        connection_context=None, op_id=1, execution_result=result
    )
    assert server.send_message.called


@pytest.mark.asyncio
async def test_resolver_with_promise(server):
    server.send_message = AsyncMock()
    result = mock.Mock()
    result.data = {"test": [1, promise.Promise(lambda resolve, reject: resolve(2))]}
    result.errors = None
    await server.send_execution_result(
        connection_context=None, op_id=1, execution_result=result
    )
    assert server.send_message.called
    assert result.data == {"test": [1, 2]}


async def test_resolver_with_nested_promise(server):
    server.send_message = AsyncMock()
    result = mock.Mock()
    inner = promise.Promise(lambda resolve, reject: resolve(2))
    outer = promise.Promise(lambda resolve, reject: resolve({"in": inner}))
    result.data = {"test": [1, outer]}
    result.errors = None
    await server.send_execution_result(
        connection_context=None, op_id=1, execution_result=result
    )
    assert server.send_message.called
    assert result.data == {"test": [1, {"in": 2}]}



from channels import __version__ as channels_version
from channels.routing import ProtocolTypeRouter, URLRouter
from channels.sessions import SessionMiddlewareStack
from django.utils.version import get_version_tuple
from django.apps import apps
from django.urls import path
from .consumers import GraphQLSubscriptionConsumer

if apps.is_installed("django.contrib.auth"):
    from channels.auth import AuthMiddlewareStack
else:
    AuthMiddlewareStack = None


channels_version_tuple = get_version_tuple(channels_version)


if channels_version_tuple > (3, 0, 0):
    websocket_urlpatterns = [
        path("subscriptions", GraphQLSubscriptionConsumer.as_asgi())
    ]
else:
    websocket_urlpatterns = [path("subscriptions", GraphQLSubscriptionConsumer)]

application = ProtocolTypeRouter({"websocket": URLRouter(websocket_urlpatterns)})

session_application = ProtocolTypeRouter(
    {"websocket": SessionMiddlewareStack(URLRouter(websocket_urlpatterns))}
)

if AuthMiddlewareStack:
    auth_application = ProtocolTypeRouter(
        {"websocket": AuthMiddlewareStack(URLRouter(websocket_urlpatterns))}
    )



import random
import asyncio
import graphene


class Query(graphene.ObjectType):
    base = graphene.String()


class RandomType(graphene.ObjectType):
    seconds = graphene.Int()
    random_int = graphene.Int()


class Subscription(graphene.ObjectType):
    count_seconds = graphene.Float(up_to=graphene.Int())
    random_int = graphene.Field(RandomType)

    async def resolve_count_seconds(root, info, up_to=5):
        for i in range(up_to):
            print("YIELD SECOND", i)
            yield i
            await asyncio.sleep(1.0)
        yield up_to

    async def resolve_random_int(root, info):
        i = 0
        while True:
            yield RandomType(seconds=i, random_int=random.randint(0, 500))
            await asyncio.sleep(1.0)
            i += 1


schema = graphene.Schema(query=Query, subscription=Subscription)



from __future__ import absolute_import

import json

from .base import (
    BaseConnectionContext,
    ConnectionClosedException,
)
from .base_sync import BaseSyncSubscriptionServer


class GeventConnectionContext(BaseConnectionContext):
    def receive(self):
        msg = self.ws.receive()
        return msg

    def send(self, data):
        if self.closed:
            return
        self.ws.send(json.dumps(data))

    @property
    def closed(self):
        return self.ws.closed

    def close(self, code):
        self.ws.close(code)


class GeventSubscriptionServer(BaseSyncSubscriptionServer):
    def handle(self, ws, request_context=None):
        connection_context = GeventConnectionContext(ws, request_context)
        self.on_open(connection_context)
        while True:
            try:
                if connection_context.closed:
                    raise ConnectionClosedException()
                message = connection_context.receive()
            except ConnectionClosedException:
                self.on_close(connection_context)
                return
            self.on_message(connection_context, message)



import json

from channels.generic.websocket import AsyncJsonWebsocketConsumer

from ..constants import WS_PROTOCOL
from .subscriptions import subscription_server


class GraphQLSubscriptionConsumer(AsyncJsonWebsocketConsumer):
    async def connect(self):
        self.connection_context = None
        protocol = None

        for p in ["graphql-ws", "graphql-transport-ws"]:
            if p in self.scope["subprotocols"]:
                protocol = p
                break

        if protocol is not None:
            self.connection_context = await subscription_server.handle(
                ws=self, request_context=self.scope
            )
            await self.accept(subprotocol=protocol)
        else:
            await self.close()

    async def disconnect(self, code):
        if self.connection_context:
            self.connection_context.socket_closed = True
            await subscription_server.on_close(self.connection_context)

    async def receive_json(self, content):
        subscription_server.on_message(self.connection_context, content)

    @classmethod
    async def encode_json(cls, content):
        return json.dumps(content)



from graphene_django.settings import graphene_settings
from ..base_async import BaseAsyncConnectionContext, BaseAsyncSubscriptionServer
from ..observable_aiter import setup_observable_extension

setup_observable_extension()


class ChannelsConnectionContext(BaseAsyncConnectionContext):
    def __init__(self, *args, **kwargs):
        super(ChannelsConnectionContext, self).__init__(*args, **kwargs)
        self.socket_closed = False

    async def send(self, data):
        if self.closed:
            return
        await self.ws.send_json(data)

    @property
    def closed(self):
        return self.socket_closed

    async def close(self, code):
        await self.ws.close(code=code)

    async def receive(self, code):
        """
        Unused, as the django consumer handles receiving messages and passes
        them straight to ChannelsSubscriptionServer.on_message.
        """


class ChannelsSubscriptionServer(BaseAsyncSubscriptionServer):
    async def handle(self, ws, request_context=None):
        connection_context = ChannelsConnectionContext(ws, request_context)
        await self.on_open(connection_context)
        return connection_context


subscription_server = ChannelsSubscriptionServer(schema=graphene_settings.SCHEMA)



import asyncio

import graphene
from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer

channel_layer = get_channel_layer()


class Query(graphene.ObjectType):
    hello = graphene.String()

    def resolve_hello(self, info, **kwargs):
        return "world"


class TestMessageMutation(graphene.Mutation):
    class Arguments:
        input_text = graphene.String()

    output_text = graphene.String()

    def mutate(self, info, input_text):
        async_to_sync(channel_layer.group_send)("new_message", {"data": input_text})
        return TestMessageMutation(output_text=input_text)


class Mutations(graphene.ObjectType):
    test_message = TestMessageMutation.Field()


class Subscription(graphene.ObjectType):
    count_seconds = graphene.Int(up_to=graphene.Int())
    new_message = graphene.String()

    async def resolve_count_seconds(self, info, up_to=5):
        i = 1
        while i <= up_to:
            yield str(i)
            await asyncio.sleep(1)
            i += 1

    async def resolve_new_message(self, info):
        channel_name = await channel_layer.new_channel()
        await channel_layer.group_add("new_message", channel_name)
        try:
            while True:
                message = await channel_layer.receive(channel_name)
                yield message["data"]
        finally:
            await channel_layer.group_discard("new_message", channel_name)


schema = graphene.Schema(query=Query, mutation=Mutations, subscription=Subscription)



try:
    from unittest import mock
except ImportError:
    import mock

import json

import pytest

from graphql_ws_apollo import base
from graphql_ws_apollo.base_sync import SubscriptionObserver


def test_not_implemented():
    server = base.BaseSubscriptionServer(schema=None)
    with pytest.raises(NotImplementedError):
        server.on_connection_init(connection_context=None, op_id=1, payload={})
    with pytest.raises(NotImplementedError):
        server.on_open(connection_context=None)


def test_on_stop():
    server = base.BaseSubscriptionServer(schema=None)
    context = mock.Mock()
    server.on_stop(connection_context=context, op_id=1)
    context.unsubscribe.assert_called_with(1)


def test_terminate():
    server = base.BaseSubscriptionServer(schema=None)

    context = mock.Mock()
    server.on_connection_terminate(connection_context=context, op_id=1)
    context.close.assert_called_with(1011)


def test_send_error():
    server = base.BaseSubscriptionServer(schema=None)
    context = mock.Mock()
    server.send_error(connection_context=context, op_id=1, error="test error")
    context.send.assert_called_with(
        {"id": 1, "type": "error", "payload": {"message": "test error"}}
    )


def test_message():
    server = base.BaseSubscriptionServer(schema=None)
    server.process_message = mock.Mock()
    context = mock.Mock()
    msg = {"id": 1, "type": base.GQL_CONNECTION_INIT, "payload": ""}
    server.on_message(context, msg)
    server.process_message.assert_called_with(context, msg)


def test_message_str():
    server = base.BaseSubscriptionServer(schema=None)
    server.process_message = mock.Mock()
    context = mock.Mock()
    msg = {"id": 1, "type": base.GQL_CONNECTION_INIT, "payload": ""}
    server.on_message(context, json.dumps(msg))
    server.process_message.assert_called_with(context, msg)


def test_message_invalid():
    server = base.BaseSubscriptionServer(schema=None)
    server.send_error = mock.Mock()
    server.on_message(connection_context=None, message="'not-json")
    assert server.send_error.called


def test_context_operations():
    ws = mock.Mock()
    context = base.BaseConnectionContext(ws)
    assert not context.has_operation(1)
    context.register_operation(1, None)
    assert context.has_operation(1)
    context.remove_operation(1)
    assert not context.has_operation(1)
    # Removing a non-existant operation fails silently.
    context.remove_operation(999)


def test_observer_data():
    ws = mock.Mock()
    context = base.BaseConnectionContext(ws)
    send_result, send_error, send_message = mock.Mock(), mock.Mock(), mock.Mock()
    observer = SubscriptionObserver(
        connection_context=context,
        op_id=1,
        send_execution_result=send_result,
        send_error=send_error,
        send_message=send_message,
    )
    observer.on_next("data")
    assert send_result.called
    assert not send_error.called


def test_observer_exception():
    ws = mock.Mock()
    context = base.BaseConnectionContext(ws)
    send_result, send_error, send_message = mock.Mock(), mock.Mock(), mock.Mock()
    observer = SubscriptionObserver(
        connection_context=context,
        op_id=1,
        send_execution_result=send_result,
        send_error=send_error,
        send_message=send_message,
    )
    observer.on_next(TypeError("some bad message"))
    assert send_error.called
    assert not send_result.called



import random
import graphene
from rx import Observable


class Query(graphene.ObjectType):
    base = graphene.String()


class RandomType(graphene.ObjectType):
    seconds = graphene.Int()
    random_int = graphene.Int()


class Subscription(graphene.ObjectType):

    count_seconds = graphene.Int(up_to=graphene.Int())

    random_int = graphene.Field(RandomType)

    def resolve_count_seconds(root, info, up_to=5):
        return (
            Observable.interval(1000)
            .map(lambda i: "{0}".format(i))
            .take_while(lambda i: int(i) <= up_to)
        )

    def resolve_random_int(root, info):
        return Observable.interval(1000).map(
            lambda i: RandomType(seconds=i, random_int=random.randint(0, 500))
        )


schema = graphene.Schema(query=Query, subscription=Subscription)



"""
Django settings for django_channels2 project.
"""
SECRET_KEY = "0%1c709jhmggqhk&=tci06iy+%jedfxpcoai69jd8wjzm+k2f0"
DEBUG = True


INSTALLED_APPS = ["channels", "graphql_ws_apollo.django", "graphene_django"]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
            ]
        },
    }
]

MIDDLEWARE = [
    'django.middleware.common.CommonMiddleware',
]

ROOT_URLCONF = "django_channels2.urls"
ASGI_APPLICATION = "graphql_ws_apollo.django.routing.application"


CHANNEL_LAYERS = {"default": {"BACKEND": "channels.layers.InMemoryChannelLayer"}}
GRAPHENE = {"MIDDLEWARE": [], "SCHEMA": "django_channels2.schema.schema"}



from graphql_ws_apollo.websockets_lib import WsLibSubscriptionServer
from graphql.execution.executors.asyncio import AsyncioExecutor
from sanic import Sanic, response
from sanic_graphql import GraphQLView
from schema import schema
from template import render_graphiql

app = Sanic(__name__)


@app.listener("before_server_start")
def init_graphql(app, loop):
    app.add_route(
        GraphQLView.as_view(schema=schema, executor=AsyncioExecutor(loop=loop)),
        "/graphql",
    )


@app.route("/graphiql")
async def graphiql_view(request):
    return response.html(render_graphiql())


subscription_server = WsLibSubscriptionServer(schema)


@app.websocket("/subscriptions", subprotocols=["graphql-ws"])
async def subscriptions(request, ws):
    await subscription_server.handle(ws)
    return ws


app.run(host="0.0.0.0", port=8000)



from __future__ import unicode_literals

import json

import django
import mock
from channels import Channel
from channels.test import ChannelTestCase
from django.conf import settings
from django.core.management import call_command

settings.configure(
    CHANNEL_LAYERS={
        "default": {
            "BACKEND": "asgiref.inmemory.ChannelLayer",
            "ROUTING": "tests.django_routing.channel_routing",
        },
    },
    INSTALLED_APPS=[
        "django.contrib.sessions",
        "django.contrib.contenttypes",
        "django.contrib.auth",
    ],
    DATABASES={"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}},
)
django.setup()

from graphql_ws_apollo.constants import GQL_CONNECTION_ACK, GQL_CONNECTION_INIT  # noqa: E402
from graphql_ws_apollo.django_channels import (  # noqa: E402
    DjangoChannelConnectionContext,
    DjangoChannelSubscriptionServer,
    GraphQLSubscriptionConsumer,
)


class TestConnectionContext:
    def test_send(self):
        msg = mock.Mock()
        connection_context = DjangoChannelConnectionContext(message=msg)
        connection_context.send("test")
        msg.reply_channel.send.assert_called_with({"text": '"test"'})

    def test_close(self):
        msg = mock.Mock()
        connection_context = DjangoChannelConnectionContext(message=msg)
        connection_context.close(123)
        msg.reply_channel.send.assert_called_with({"close": True, "text": 123})


def test_subscription_server_smoke():
    DjangoChannelSubscriptionServer(schema=None)


class TestConsumer(ChannelTestCase):
    def test_connect(self):
        call_command("migrate")
        Channel("websocket.receive").send(
            {
                "path": "/graphql",
                "order": 0,
                "reply_channel": "websocket.receive",
                "text": json.dumps({"type": GQL_CONNECTION_INIT, "id": 1}),
            }
        )
        message = self.get_next_message("websocket.receive", require=True)
        GraphQLSubscriptionConsumer(message)
        result = self.get_next_message("websocket.receive", require=True)
        result_content = json.loads(result.content["text"])
        assert result_content == {"type": GQL_CONNECTION_ACK}



import json

from aiohttp import web
from schema import schema
from graphql import format_error
from graphql_ws_apollo.aiohttp import AiohttpSubscriptionServer

from template import render_graphiql


async def graphql_view(request):
    payload = await request.json()
    response = await schema.execute(payload.get("query", ""), return_promise=True)
    data = {}
    if response.errors:
        data["errors"] = [format_error(e) for e in response.errors]
    if response.data:
        data["data"] = response.data
    jsondata = json.dumps(data,)
    return web.Response(text=jsondata, headers={"Content-Type": "application/json"})


async def graphiql_view(request):
    return web.Response(text=render_graphiql(), headers={"Content-Type": "text/html"})


subscription_server = AiohttpSubscriptionServer(schema)


async def subscriptions(request):
    ws = web.WebSocketResponse(protocols=("graphql-ws",))
    await ws.prepare(request)

    await subscription_server.handle(ws)
    return ws


app = web.Application()
app.router.add_get("/subscriptions", subscriptions)
app.router.add_get("/graphiql", graphiql_view)
app.router.add_get("/graphql", graphql_view)
app.router.add_post("/graphql", graphql_view)

web.run_app(app, port=8000)



import json
from asyncio import shield

from websockets import ConnectionClosed

from .base import ConnectionClosedException
from .base_async import BaseAsyncConnectionContext, BaseAsyncSubscriptionServer


class WsLibConnectionContext(BaseAsyncConnectionContext):
    async def receive(self):
        try:
            msg = await self.ws.recv()
            return msg
        except ConnectionClosed:
            raise ConnectionClosedException()

    async def send(self, data):
        if self.closed:
            return
        await self.ws.send(json.dumps(data))

    @property
    def closed(self):
        return self.ws.open is False

    async def close(self, code):
        await self.ws.close(code)


class WsLibSubscriptionServer(BaseAsyncSubscriptionServer):
    async def _handle(self, ws, request_context):
        connection_context = WsLibConnectionContext(ws, request_context)
        await self.on_open(connection_context)
        while True:
            try:
                if connection_context.closed:
                    raise ConnectionClosedException()
                message = await connection_context.receive()
            except ConnectionClosedException:
                break

            self.on_message(connection_context, message)
        await self.on_close(connection_context)

    async def handle(self, ws, request_context=None):
        await shield(self._handle(ws, request_context), loop=self.loop)



try:
    from unittest import mock
except ImportError:
    import mock

from graphql_ws_apollo.gevent import GeventConnectionContext, GeventSubscriptionServer


class TestConnectionContext:
    def test_receive(self):
        ws = mock.Mock()
        connection_context = GeventConnectionContext(ws=ws)
        connection_context.receive()
        ws.receive.assert_called()

    def test_send(self):
        ws = mock.Mock()
        ws.closed = False
        connection_context = GeventConnectionContext(ws=ws)
        connection_context.send({"text": "test"})
        ws.send.assert_called_with('{"text": "test"}')

    def test_send_closed(self):
        ws = mock.Mock()
        ws.closed = True
        connection_context = GeventConnectionContext(ws=ws)
        connection_context.send("test")
        assert not ws.send.called

    def test_close(self):
        ws = mock.Mock()
        connection_context = GeventConnectionContext(ws=ws)
        connection_context.close(123)
        ws.close.assert_called_with(123)


def test_subscription_server_smoke():
    GeventSubscriptionServer(schema=None)



import json

try:
    # Channels version > 1 renamed the websockets module to websocket.
    from channels.generic.websockets import JsonWebsocketConsumer
except ImportError:
    from channels.generic.websocket import JsonWebsocketConsumer
from graphene_django.settings import graphene_settings

from .base import BaseConnectionContext
from .base_sync import BaseSyncSubscriptionServer


class DjangoChannelConnectionContext(BaseConnectionContext):
    def __init__(self, message):
        super(DjangoChannelConnectionContext, self).__init__(
            message.reply_channel,
            request_context={"user": message.user, "session": message.http_session},
        )

    def send(self, data):
        self.ws.send({"text": json.dumps(data)})

    def close(self, reason):
        data = {"close": True, "text": reason}
        self.ws.send(data)


class DjangoChannelSubscriptionServer(BaseSyncSubscriptionServer):
    def handle(self, message, connection_context):
        self.on_message(connection_context, message)


subscription_server = DjangoChannelSubscriptionServer(graphene_settings.SCHEMA)


class GraphQLSubscriptionConsumer(JsonWebsocketConsumer):
    http_user_and_session = True
    strict_ordering = True

    def connect(self, message, **kwargs):
        message.reply_channel.send({"accept": True})

    def receive(self, content, **kwargs):
        """
        Called when a message is received with either text or bytes
        filled out.
        """
        context = DjangoChannelConnectionContext(self.message)
        subscription_server.on_open(context)
        subscription_server.handle(content, context)



GRAPHQL_WS = "graphql-ws-apollo"
WS_PROTOCOL = GRAPHQL_WS

GQL_CONNECTION_INIT = "connection_init"  # Client -> Server
GQL_CONNECTION_ACK = "connection_ack"  # Server -> Client
GQL_CONNECTION_ERROR = "connection_error"  # Server -> Client

# NOTE: This one here don't follow the standard due to connection optimization
GQL_CONNECTION_TERMINATE = "connection_terminate"  # Client -> Server
GQL_CONNECTION_KEEP_ALIVE = "ka"  # Server -> Client

GQL_START = "start"  # Client -> Server
GQL_START_APOLLO = "subscribe" # Client -> Server
GQL_DATA = "data"  # Server -> Client
GQL_DATA_APOLLO = "next"  # Server -> Client

GQL_ERROR = "error"  # Server -> Client
GQL_COMPLETE = "complete"  # Server -> Client
GQL_STOP = "stop"  # Client -> Server
GQL_STOP_APOLLO = "complete" # Client -> Server



from graphql.execution.executors.sync import SyncExecutor
from rx import Observable, Observer

from .base import BaseSubscriptionServer
from .constants import GQL_COMPLETE, GQL_CONNECTION_ACK, GQL_CONNECTION_ERROR, GQL_DATA


class BaseSyncSubscriptionServer(BaseSubscriptionServer):
    graphql_executor = SyncExecutor

    def on_operation_complete(self, connection_context, op_id):
        pass

    def handle(self, ws, request_context=None):
        raise NotImplementedError("handle method not implemented")

    def on_open(self, connection_context):
        pass

    def on_connect(self, connection_context, payload):
        pass

    def on_connection_init(self, connection_context, op_id, payload):
        try:
            self.on_connect(connection_context, payload)
            self.send_message(connection_context, op_type=GQL_CONNECTION_ACK)

        except Exception as e:
            self.send_error(connection_context, op_id, e, GQL_CONNECTION_ERROR)
            connection_context.close(1011)

    def on_start(self, connection_context, op_id, params, data_op_type=GQL_DATA):
        # Attempt to unsubscribe first in case we already have a subscription
        # with this id.
        connection_context.unsubscribe(op_id)
        try:
            execution_result = self.execute(params)
            assert isinstance(
                execution_result, Observable
            ), "A subscription must return an observable"
            disposable = execution_result.subscribe(
                SubscriptionObserver(
                    connection_context,
                    op_id,
                    self.send_execution_result,
                    self.send_error,
                    self.send_message,
                    data_op_type,
                )
            )
            connection_context.register_operation(op_id, disposable)

        except Exception as e:
            self.send_error(connection_context, op_id, e)
            self.send_message(connection_context, op_id, GQL_COMPLETE)


class SubscriptionObserver(Observer):
    def __init__(
        self, connection_context, op_id, send_execution_result, send_error, send_message, data_op_type=GQL_DATA
    ):
        self.connection_context = connection_context
        self.op_id = op_id
        self.send_execution_result = send_execution_result
        self.send_error = send_error
        self.send_message = send_message
        self.data_op_type = data_op_type

    def on_next(self, value):
        if isinstance(value, Exception):
            send_method = self.send_error
        else:
            send_method = self.send_execution_result
        send_method(self.connection_context, self.op_id, value, self.data_op_type)

    def on_completed(self):
        self.send_message(self.connection_context, self.op_id, GQL_COMPLETE)
        self.connection_context.remove_operation(self.op_id)

    def on_error(self, error):
        self.send_error(self.connection_context, self.op_id, error)
        self.on_completed()



import json
from collections import OrderedDict

from graphql import format_error, graphql

from .constants import (
    GQL_CONNECTION_ERROR,
    GQL_CONNECTION_INIT,
    GQL_CONNECTION_TERMINATE,
    GQL_DATA,
    GQL_DATA_APOLLO,
    GQL_ERROR,
    GQL_START,
    GQL_START_APOLLO,
    GQL_STOP,
    GQL_STOP_APOLLO,
)


class ConnectionClosedException(Exception):
    pass


class BaseConnectionContext(object):
    def __init__(self, ws, request_context=None):
        self.ws = ws
        self.operations = {}
        self.request_context = request_context

    def has_operation(self, op_id):
        return op_id in self.operations

    def register_operation(self, op_id, async_iterator):
        self.operations[op_id] = async_iterator

    def get_operation(self, op_id):
        return self.operations[op_id]

    def remove_operation(self, op_id):
        try:
            return self.operations.pop(op_id)
        except KeyError:
            return

    def unsubscribe(self, op_id):
        async_iterator = self.remove_operation(op_id)
        if hasattr(async_iterator, "dispose"):
            async_iterator.dispose()
        return async_iterator

    def unsubscribe_all(self):
        for op_id in list(self.operations):
            self.unsubscribe(op_id)

    def receive(self):
        raise NotImplementedError("receive method not implemented")

    def send(self, data):
        raise NotImplementedError("send method not implemented")

    @property
    def closed(self):
        raise NotImplementedError("closed property not implemented")

    def close(self, code):
        raise NotImplementedError("close method not implemented")


class BaseSubscriptionServer(object):
    graphql_executor = None

    def __init__(self, schema, keep_alive=True):
        self.schema = schema
        self.keep_alive = keep_alive

    def execute(self, params):
        return graphql(self.schema, **dict(params, allow_subscriptions=True))

    def process_message(self, connection_context, parsed_message):
        op_id = parsed_message.get("id")
        op_type = parsed_message.get("type")
        payload = parsed_message.get("payload")

        if op_type == GQL_CONNECTION_INIT:
            return self.on_connection_init(connection_context, op_id, payload)

        elif op_type == GQL_CONNECTION_TERMINATE:
            return self.on_connection_terminate(connection_context, op_id)

        elif op_type == GQL_START or op_type == GQL_START_APOLLO:
            assert isinstance(payload, dict), "The payload must be a dict"
            params = self.get_graphql_params(connection_context, payload)
            data_op_type = GQL_DATA if op_type == GQL_START else GQL_DATA_APOLLO
            return self.on_start(connection_context, op_id, params, data_op_type)

        elif op_type == GQL_STOP or op_type == GQL_STOP_APOLLO:
            return self.on_stop(connection_context, op_id)

        else:
            return self.send_error(
                connection_context,
                op_id,
                Exception("Invalid message type: {}.".format(op_type)),
            )

    def on_connection_init(self, connection_context, op_id, payload):
        raise NotImplementedError("on_connection_init method not implemented")

    def on_connection_terminate(self, connection_context, op_id):
        return connection_context.close(1011)

    def get_graphql_params(self, connection_context, payload):
        context = payload.get("context", connection_context.request_context)
        return {
            "request_string": payload.get("query"),
            "variable_values": payload.get("variables"),
            "operation_name": payload.get("operationName"),
            "context_value": context,
            "executor": self.graphql_executor(),
        }

    def on_open(self, connection_context):
        raise NotImplementedError("on_open method not implemented")

    def on_stop(self, connection_context, op_id):
        return connection_context.unsubscribe(op_id)

    def on_close(self, connection_context):
        return connection_context.unsubscribe_all()

    def send_message(self, connection_context, op_id=None, op_type=None, payload=None):
        if op_id is None or connection_context.has_operation(op_id):
            message = self.build_message(op_id, op_type, payload)
            return connection_context.send(message)

    def build_message(self, id, op_type, payload):
        message = {}
        if id is not None:
            message["id"] = id
        if op_type is not None:
            message["type"] = op_type
        if payload is not None:
            message["payload"] = payload
        assert message, "You need to send at least one thing"
        return message

    def send_execution_result(self, connection_context, op_id, execution_result, data_op_type=GQL_DATA):
        result = self.execution_result_to_dict(execution_result)
        return self.send_message(connection_context, op_id, data_op_type, result)

    def execution_result_to_dict(self, execution_result):
        result = OrderedDict()
        if execution_result.data:
            result["data"] = execution_result.data
        if execution_result.errors:
            result["errors"] = [
                format_error(error) for error in execution_result.errors
            ]
        return result

    def send_error(self, connection_context, op_id, error, error_type=None):
        if error_type is None:
            error_type = GQL_ERROR

        assert error_type in [GQL_CONNECTION_ERROR, GQL_ERROR], (
            "error_type should be one of the allowed error messages"
            " GQL_CONNECTION_ERROR or GQL_ERROR"
        )

        error_payload = {"message": str(error)}

        return self.send_message(connection_context, op_id, error_type, error_payload)

    def on_message(self, connection_context, message):
        try:
            if not isinstance(message, dict):
                parsed_message = json.loads(message)
                assert isinstance(parsed_message, dict), "Payload must be an object."
            else:
                parsed_message = message
        except Exception as e:
            return self.send_error(connection_context, None, e)

        return self.process_message(connection_context, parsed_message)



"""django_subscriptions URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/1.11/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  url(r'^$', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  url(r'^$', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.conf.urls import url, include
    2. Add a URL to urlpatterns:  url(r'^blog/', include('blog.urls'))
"""
from django.conf.urls import url
from django.contrib import admin
from .template import render_graphiql
from django.http import HttpResponse

from graphene_django.views import GraphQLView
from django.views.decorators.csrf import csrf_exempt

from channels.routing import route_class
from graphql_ws_apollo.django_channels import GraphQLSubscriptionConsumer


def graphiql(request):
    response = HttpResponse(content=render_graphiql())
    return response


urlpatterns = [
    url(r"^admin/", admin.site.urls),
    url(r"^graphiql/", graphiql),
    url(r"^graphql", csrf_exempt(GraphQLView.as_view(graphiql=True))),
]

channel_routing = [
    route_class(GraphQLSubscriptionConsumer, path=r"^/subscriptions"),
]
